@@ -4,11 +4,24 @@ package Hash::Merge;
 
 #=============================================================================
 #
-# $Id: Merge.pm,v 0.06 2001/11/10 03:30:34 mneylon Exp $
-# $Revision: 0.06 $
+# $Id: Merge.pm,v 0.07 2002/02/19 00:21:27 mneylon Exp $
+# $Revision: 0.07 $
 # $Author: mneylon $
-# $Date: 2001/11/10 03:30:34 $
+# $Date: 2002/02/19 00:21:27 $
 # $Log: Merge.pm,v $
+# Revision 0.07  2002/02/19 00:21:27  mneylon
+# Fixed problem with ActiveState Perl's Clone.pm implementation.
+# Fixed typo in POD.
+# Fixed formatting of code in general.
+#
+# Revision 0.06.01.2  2002/02/17 03:18:20  mneylon
+# Fixed problem with ActiveState Perl's Clone.pm implementation.
+# Fixed typo in POD.
+# Fixed formatting of code in general.
+#
+# Revision 0.06.01.1  2002/02/17 02:48:54  mneylon
+# Branched version.
+#
 # Revision 0.06  2001/11/10 03:30:34  mneylon
 # Version 0.06 release (and more CVS fixes)
 #
@@ -50,81 +63,84 @@ use strict;
 use Clone qw(clone);
 
 BEGIN {
-    use Exporter   ();
-    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
-    $VERSION     = sprintf( "%d.%02d", q($Revision: 0.06 $) =~ /\s(\d+)\.(\d+)/ );
-    @ISA         = qw(Exporter);
-    @EXPORT      = qw();
-	@EXPORT_OK   = qw( merge _hashify _merge_hashes );
-    %EXPORT_TAGS = ( custom => [ qw( _hashify _merge_hashes )] );
+  use Exporter   ();
+  use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK 
+		    %EXPORT_TAGS $CLONE_SUPPORT);
+   $VERSION     = sprintf( "%d.%02d", q($Revision: 0.07 $) =~ /\s(\d+)\.(\d+)/ );
+  @ISA         = qw(Exporter);
+  @EXPORT      = qw();
+  @EXPORT_OK   = qw( merge _hashify _merge_hashes );
+  %EXPORT_TAGS  = ( custom => [ qw( _hashify _merge_hashes )] );
+  $CLONE_SUPPORT = ( $Clone::VERSION > 0.09 );
+  
 }
 
 my %left_precedent = (
-	SCALAR => {
-		SCALAR => sub { $_[0] },
-		ARRAY  => sub { $_[0] },
-		HASH   => sub { $_[0] } },
-	ARRAY => {
-		SCALAR => sub { [ @{$_[0]}, $_[1] ] },
-		ARRAY  => sub { [ @{$_[0]}, @{$_[1]} ] },
-		HASH   => sub { [ @{$_[0]}, values %{$_[1]} ] } },
-	HASH => {
-		SCALAR => sub { $_[0] },
-		ARRAY  => sub { $_[0] },
-		HASH   => sub { _merge_hashes( $_[0], $_[1] ) } }
-);
+		      SCALAR => {
+				 SCALAR => sub { $_[0] },
+				 ARRAY  => sub { $_[0] },
+				 HASH   => sub { $_[0] } },
+		      ARRAY => {
+				SCALAR => sub { [ @{$_[0]}, $_[1] ] },
+				ARRAY  => sub { [ @{$_[0]}, @{$_[1]} ] },
+				HASH   => sub { [ @{$_[0]}, values %{$_[1]} ] } },
+		      HASH => {
+			       SCALAR => sub { $_[0] },
+			       ARRAY  => sub { $_[0] },
+			       HASH   => sub { _merge_hashes( $_[0], $_[1] ) } }
+		     );
 
 my %right_precedent = (
-	SCALAR => {
-		SCALAR => sub { $_[1] },
-		ARRAY  => sub { [ $_[0], @{$_[1]} ] },
-		HASH   => sub { $_[1] } },
-	ARRAY => {
-		SCALAR => sub { $_[1] },
-		ARRAY  => sub { [ @{$_[0]}, @{$_[1]} ] },
-		HASH   => sub { $_[1] } },
-	HASH => {
-		SCALAR => sub { $_[1] },
-		ARRAY  => sub { [ values %{$_[0]}, @{$_[1]} ] },
-		HASH   => sub { _merge_hashes( $_[0], $_[1] ) } }
-);
+		       SCALAR => {
+				  SCALAR => sub { $_[1] },
+				  ARRAY  => sub { [ $_[0], @{$_[1]} ] },
+				  HASH   => sub { $_[1] } },
+		       ARRAY => {
+				 SCALAR => sub { $_[1] },
+				 ARRAY  => sub { [ @{$_[0]}, @{$_[1]} ] },
+				 HASH   => sub { $_[1] } },
+		       HASH => {
+				SCALAR => sub { $_[1] },
+				ARRAY  => sub { [ values %{$_[0]}, @{$_[1]} ] },
+				HASH   => sub { _merge_hashes( $_[0], $_[1] ) } }
+		      );
 
 my %storage_precedent = (
-	SCALAR => {
-		SCALAR => sub { $_[0] },
-		ARRAY  => sub { [ $_[0], @{$_[1]} ] },
-		HASH   => sub { $_[1] } },
-	ARRAY => {
-		SCALAR => sub { [ @{$_[0]}, $_[1] ] },
-		ARRAY  => sub { [ @{$_[0]}, @{$_[1]} ] },
-		HASH   => sub { $_[1] } },
-	HASH => {
-		SCALAR => sub { $_[0] },
-		ARRAY  => sub { $_[0] },
-		HASH   => sub { _merge_hashes( $_[0], $_[1] ) } }
-);
+			 SCALAR => {
+				    SCALAR => sub { $_[0] },
+				    ARRAY  => sub { [ $_[0], @{$_[1]} ] },
+				    HASH   => sub { $_[1] } },
+			 ARRAY => {
+				   SCALAR => sub { [ @{$_[0]}, $_[1] ] },
+				   ARRAY  => sub { [ @{$_[0]}, @{$_[1]} ] },
+				   HASH   => sub { $_[1] } },
+			 HASH => {
+				  SCALAR => sub { $_[0] },
+				  ARRAY  => sub { $_[0] },
+				  HASH   => sub { _merge_hashes( $_[0], $_[1] ) } }
+			);
 
 my %retainment_precedent = (
-	SCALAR => {
-		SCALAR => sub { [ $_[0], $_[1] ] },
-		ARRAY  => sub { [ $_[0], @{$_[1]} ] },
-		HASH   => sub { _merge_hashes( _hashify( $_[0] ), $_[1] ) } },
-	ARRAY => {
-		SCALAR => sub { [ @{$_[0]}, $_[1] ] },
-		ARRAY  => sub { [ @{$_[0]}, @{$_[1]} ] },
-		HASH   => sub { _merge_hashes( _hashify( $_[0] ), $_[1] ) } },
-	HASH => {
-		SCALAR => sub { _merge_hashes( $_[0], _hashify( $_[1] ) ) },
-		ARRAY  => sub { _merge_hashes( $_[0], _hashify( $_[1] ) ) },
-		HASH   => sub { _merge_hashes( $_[0], $_[1] ) } }
-);
+			    SCALAR => {
+				       SCALAR => sub { [ $_[0], $_[1] ] },
+				       ARRAY  => sub { [ $_[0], @{$_[1]} ] },
+				       HASH   => sub { _merge_hashes( _hashify( $_[0] ), $_[1] ) } },
+			    ARRAY => {
+				      SCALAR => sub { [ @{$_[0]}, $_[1] ] },
+				      ARRAY  => sub { [ @{$_[0]}, @{$_[1]} ] },
+				      HASH   => sub { _merge_hashes( _hashify( $_[0] ), $_[1] ) } },
+			    HASH => {
+				     SCALAR => sub { _merge_hashes( $_[0], _hashify( $_[1] ) ) },
+				     ARRAY  => sub { _merge_hashes( $_[0], _hashify( $_[1] ) ) },
+				     HASH   => sub { _merge_hashes( $_[0], $_[1] ) } }
+			   );
 
 my %behaviors = (
-	LEFT_PRECEDENT => \%left_precedent,
-	RIGHT_PRECEDENT => \%right_precedent,
-	STORAGE_PRECEDENT => \%storage_precedent,
-	RETAINMENT_PRECEDENT => \%retainment_precedent 
-);
+		 LEFT_PRECEDENT => \%left_precedent,
+		 RIGHT_PRECEDENT => \%right_precedent,
+		 STORAGE_PRECEDENT => \%storage_precedent,
+		 RETAINMENT_PRECEDENT => \%retainment_precedent 
+		);
 
 my $merge_behavior = 'LEFT_PRECEDENT';
 my $merge_matrix = \%{ $behaviors{ $merge_behavior } };
@@ -132,131 +148,154 @@ my $merge_matrix = \%{ $behaviors{ $merge_behavior } };
 my $clone_behavior = 1;
 
 sub set_behavior {
-	my $value = uc(shift);
-	die "Behavior must be one of : " , join ' ', keys %behaviors 
-		unless exists $behaviors{ $value };
-	$merge_behavior = $value;
-	$merge_matrix = \%{ $behaviors{ $merge_behavior } };
+  my $value = uc(shift);
+  die "Behavior must be one of : " , join ' ', keys %behaviors 
+    unless exists $behaviors{ $value };
+  $merge_behavior = $value;
+  $merge_matrix = \%{ $behaviors{ $merge_behavior } };
 }
 
 sub get_behavior {
-	return $merge_behavior;
+  return $merge_behavior;
 }
 
 sub specify_behavior {
-	my $matrix = shift;
-	my $name = shift || "user defined";
-	my @required = qw ( SCALAR ARRAY HASH );
-
-	foreach my $left ( @required ) {
-		foreach my $right ( @required ) {
-			die "Behavior does not specify action for $left merging with $right"
-				unless exists $matrix->{ $left }->{ $right };
-		}
-	}
-
-	$merge_behavior = $name;
-	$merge_matrix = $matrix;
+  my $matrix = shift;
+  my $name = shift || "user defined";
+  my @required = qw ( SCALAR ARRAY HASH );
+  
+  foreach my $left ( @required ) {
+    foreach my $right ( @required ) {
+      die "Behavior does not specify action for $left merging with $right"
+	unless exists $matrix->{ $left }->{ $right };
+    }
+  }
+  
+  $merge_behavior = $name;
+  $merge_matrix = $matrix;
 }
 
 sub set_clone_behavior {
-    my $temp = shift;
-	$clone_behavior = ( $temp ) ? 1 : 0;
+  my $temp = shift;
+  $clone_behavior = ( $temp ) ? 1 : 0;
 }
 
 sub get_clone_behavior {
-	return $clone_behavior;
+  return $clone_behavior;
 }
 
 sub merge {
-	my ( $left, $right ) = ( shift, shift );
-
-	# For the general use of this module, we want to create duplicates
-	# of all data that is merged.  This behavior can be shut off, but 
-	# can mess havoc if references are used heavily.
-
-	my ( $lefttype, $righttype );
-	if ( !defined( $left ) ) {		# Perl 5.005 compatibility
-		$lefttype = 'SCALAR';
-	} elsif ( UNIVERSAL::isa( $left, 'HASH' ) ) { 
-		$lefttype = 'HASH';
-	} elsif ( UNIVERSAL::isa( $left, 'ARRAY' ) ) {
-		$lefttype = 'ARRAY';
-	} else {
-		$lefttype = 'SCALAR';
-	}
-
-	if ( !defined( $right ) ) {		# Perl 5.005 compatibility
-		$righttype = 'SCALAR';
-	} elsif ( UNIVERSAL::isa( $right, 'HASH' ) ) { 
-		$righttype = 'HASH';
-	} elsif ( UNIVERSAL::isa( $right, 'ARRAY' ) ) {
-		$righttype = 'ARRAY';
-	} else {
-		$righttype = 'SCALAR';
-	}
-	
-	if ( $clone_behavior ) {
-		$left = clone ( $left, 1 ); 
-		$right = clone ( $right, 1 );
-	}
-
-	return &{ $merge_matrix->{ $lefttype }->{ $righttype }}
-		( $left, $right );
+  my ( $left, $right ) = ( shift, shift );
+  
+  # For the general use of this module, we want to create duplicates
+  # of all data that is merged.  This behavior can be shut off, but 
+  # can mess havoc if references are used heavily.
+  
+  my ( $lefttype, $righttype );
+  if ( !defined( $left ) ) {		# Perl 5.005 compatibility
+    $lefttype = 'SCALAR';
+  } elsif ( UNIVERSAL::isa( $left, 'HASH' ) ) { 
+    $lefttype = 'HASH';
+  } elsif ( UNIVERSAL::isa( $left, 'ARRAY' ) ) {
+    $lefttype = 'ARRAY';
+  } else {
+    $lefttype = 'SCALAR';
+  }
+  
+  if ( !defined( $right ) ) {		# Perl 5.005 compatibility
+    $righttype = 'SCALAR';
+  } elsif ( UNIVERSAL::isa( $right, 'HASH' ) ) { 
+    $righttype = 'HASH';
+  } elsif ( UNIVERSAL::isa( $right, 'ARRAY' ) ) {
+    $righttype = 'ARRAY';
+  } else {
+    $righttype = 'SCALAR';
+  }
+  
+  if ( $clone_behavior ) {
+    $left = _my_clone ( $left, 1 ); 
+    $right = _my_clone ( $right, 1 );
+  }
+  
+  return &{ $merge_matrix->{ $lefttype }->{ $righttype }}
+    ( $left, $right );
 }	
 
 # This does a straight merge of hashes, delegating the merge-specific 
 # work to 'merge'
 
 sub _merge_hashes {
-	my ( $left, $right ) = ( shift, shift );
-	die "Arguments for _merge_hashes must be hash references" unless 
-		UNIVERSAL::isa( $left, 'HASH' ) && UNIVERSAL::isa( $right, 'HASH' );
-	
-	my %newhash;
-	foreach my $leftkey ( keys %$left ) {
-		if ( exists $right->{ $leftkey } ) {
-			$newhash{ $leftkey } = 
-				merge ( $left->{ $leftkey }, $right->{ $leftkey } )
-		} else {
-			$newhash{ $leftkey } = 
-				( $clone_behavior ) ? clone( $left->{ $leftkey } )
-								    : $left->{ $leftkey };
-		}
-	}
-	foreach my $rightkey ( keys %$right ) { 
-		if ( !exists $left->{ $rightkey } ) {
-			$newhash{ $rightkey } = 
-				( $clone_behavior ) ? clone( $right->{ $rightkey } )
-								    : $right->{ $rightkey };
-		}
-	}
-	return \%newhash;
+  my ( $left, $right ) = ( shift, shift );
+  die "Arguments for _merge_hashes must be hash references" unless 
+    UNIVERSAL::isa( $left, 'HASH' ) && UNIVERSAL::isa( $right, 'HASH' );
+  
+  my %newhash;
+  foreach my $leftkey ( keys %$left ) {
+    if ( exists $right->{ $leftkey } ) {
+      $newhash{ $leftkey } = 
+	merge ( $left->{ $leftkey }, $right->{ $leftkey } )
+      } else {
+	$newhash{ $leftkey } = 
+	  ( $clone_behavior ) ? _my_clone( $left->{ $leftkey } )
+	    : $left->{ $leftkey };
+      }
+  }
+  foreach my $rightkey ( keys %$right ) { 
+    if ( !exists $left->{ $rightkey } ) {
+      $newhash{ $rightkey } = 
+	( $clone_behavior ) ? _my_clone( $right->{ $rightkey } )
+	  : $right->{ $rightkey };
+    }
+  }
+  return \%newhash;
 }
 
 # Given a scalar or an array, creates a new hash where for each item in
 # the passed scalar or array, the key is equal to the value.  Returns
 # this new hash
- 
+
 sub _hashify {
-	my $arg = shift;
-	die "Arguement for _hashify must not be a HASH ref" if
-		UNIVERSAL::isa( $arg, 'HASH' );
-	
-	my %newhash;
-	if ( UNIVERSAL::isa( $arg, 'ARRAY' ) ) {
-		foreach my $item ( @$arg ) {
-		    my $suffix = 2;
-			my $name = $item;
-			while ( exists $newhash{ $name } ) {
-				$name = $item . $suffix++;
-			}
-			$newhash{ $name } = $item;
-		}
-	} else {
-		$newhash{ $arg } = $arg;
-	}
-	return \%newhash;
+  my $arg = shift;
+  die "Arguement for _hashify must not be a HASH ref" if
+    UNIVERSAL::isa( $arg, 'HASH' );
+  
+  my %newhash;
+  if ( UNIVERSAL::isa( $arg, 'ARRAY' ) ) {
+    foreach my $item ( @$arg ) {
+      my $suffix = 2;
+      my $name = $item;
+      while ( exists $newhash{ $name } ) {
+	$name = $item . $suffix++;
+      }
+      $newhash{ $name } = $item;
+    }
+  } else {
+    $newhash{ $arg } = $arg;
+  }
+  return \%newhash;
+}
+
+# This adds some checks to the clone process, to deal with problems that 
+# the current distro of ActiveState perl has (specifically, it uses 0.09
+# of Clone, which does not support the cloning of scalars).  This simply
+# wraps around clone as to prevent a scalar from being cloned via a 
+# Clone 0.09 process.  This might mean that CODEREFs and anything else
+# not a HASH or ARRAY won't be cloned.
+
+sub _my_clone {
+  my ( $arg, $depth ) = @_;
+  if ( !$CLONE_SUPPORT && 
+       !UNIVERSAL::isa( $arg, 'HASH' ) && 
+       !UNIVERSAL::isa( $arg, 'ARRAY' )) { 
+      my $var = $arg; # Forced clone
+      return $var;
+  } else {
+    if ($depth ) {
+      return clone( $arg, $depth );
+    } else {
+      return clone( $arg );
+    }
+  }
 }
 
 1;
@@ -278,7 +317,7 @@ Hash::Merge - Merges arbitrarily deep hashes into a single hash
 
   my %c = %{ merge( \%a, \%b ) };
 
-  Hash::Merge::set_behavior( RIGHT_PRECEDENCE );
+  Hash::Merge::set_behavior( RIGHT_PRECEDENT );
 
   # This is the same as above
 
@@ -442,13 +481,29 @@ the right as $b (these are either scalars or appropriate references)
 This will not handle self-referencing/recursion within hashes well.  
 Plans for a future version include incorporate deep recursion protection.
 
+As of Feb 16, 2002, ActiveState Perl's PPM of Clone.pm is only at
+0.09.  This version does not support the cloning of scalars if passed
+to the function.  This is fixed by 0.10 (and currently, Clone.pm is at
+0.13).  So while most other users can upgrade their Clone.pm
+appropriately (and I could put this as a requirement into the
+Makefile.PL), those using ActiveState would lose out on the ability to
+use this module.  (Clone.pm is not pure perl, so it's not simply a
+matter of moving the newer file into place).  Thus, for the time
+being, a check is done at the start of loading of this module to see
+if a newer version of clone is around.  Then, all cloning calls have
+been wrapped in the internal _my_clone function to block any scalar
+clones if Clone.pm is too old.  However, this also prevents the
+cloning of anything that isn't a hash or array under the same
+conditions.  Once ActiveState updates their Clone, I'll remove this 
+wrapper.
+
 =head1 AUTHOR
 
 Michael K. Neylon E<lt>mneylon-pm@masemware.comE<gt>
 
 =head1 COPYRIGHT
 
-Copyright (c) 2001 Michael K. Neylon. All rights reserved.
+Copyright (c) 2001,2002 Michael K. Neylon. All rights reserved.
 
 This library is free software.  You can redistribute it and/or modify it 
 under the same terms as Perl itself.
@@ -456,6 +511,19 @@ under the same terms as Perl itself.
 =head1 HISTORY
 
 	$Log: Merge.pm,v $
+	Revision 0.07  2002/02/19 00:21:27  mneylon
+	Fixed problem with ActiveState Perl's Clone.pm implementation.
+	Fixed typo in POD.
+	Fixed formatting of code in general.
+	
+	Revision 0.06.01.2  2002/02/17 03:18:20  mneylon
+	Fixed problem with ActiveState Perl's Clone.pm implementation.
+	Fixed typo in POD.
+	Fixed formatting of code in general.
+	
+	Revision 0.06.01.1  2002/02/17 02:48:54  mneylon
+	Branched version.
+	
 	Revision 0.06  2001/11/10 03:30:34  mneylon
 	Version 0.06 release (and more CVS fixes)
 	
